import codecs
import os
import xml.etree.ElementTree as ET

from .basemodel import BaseModel
from ..unicodecsv import UnicodeDictReader

class NestedMKModel(BaseModel):

    def __init__(self, model_config, global_config):

        BaseModel.__init__(self, model_config, global_config)

    def add_state(self, state):
        BaseModel.add_state(self, state)
        for f in self.features:
            fname = "%s:%s" % (self.name, f)
            alpha = ET.SubElement(state, "parameter", {"id":"nested_alpha.s:%s" % fname, "lower":"1.0E-4", "name":"stateNode", "upper":"1.0E4"})
            alpha.text="0.5"

    def add_sitemodel(self, distribution, feature, fname):

            # Sitemodel
            if self.rate_variation:
                mr = "@featureClockRate:%s" % fname
            else:
                mr = "1.0"
            sitemodel = ET.SubElement(distribution, "siteModel", {"id":"geoSiteModel.%s"%fname,"spec":"SiteModel", "mutationRate":mr,"shape":"1","proportionInvariant":"0"})

            substmodel = ET.SubElement(sitemodel, "substModel",{"id":"nestedmk.s:%s"%fname,"spec":"NestedLewisMK","alpha":"@nested_alpha.s:%s"%fname, "stateNumber":"%d" % self.valuecounts[feature]})
            if self.frequencies == "empirical":
                if self.pruned:
                    freq = ET.SubElement(substmodel,"frequencies",{"id":"featurefreqs.s:%s"%fname,"spec":"Frequencies", "data":"@%s.filt"%fname})
                else:
                    freq = ET.SubElement(substmodel,"frequencies",{"id":"featurefreqs.s:%s"%fname,"spec":"Frequencies", "data":"@%s"%fname})

    def add_operators(self, run):
        BaseModel.add_operators(self, run)
        for f in self.features:
            fname = "%s:%s" % (self.name, f)
            ET.SubElement(run, "operator", {"id":"nested_alpha_scaler.s:%s" % fname, "spec":"ScaleOperator","parameter":"@nested_alpha.s:%s" % fname,"scaleFactor":"0.75","weight":"0.1"})

    def add_param_logs(self, logger):
        BaseModel.add_param_logs(self, logger)
        for f in self.features:
            fname = "%s:%s" % (self.name, f)
            ET.SubElement(logger,"log",{"idref":"nested_alpha.s:%s" % fname})



import codecs
import os
import xml.etree.ElementTree as ET

from .basemodel import BaseModel
from ..unicodecsv import UnicodeDictReader

class OrdinalModel(BaseModel):

    def __init__(self, model_config, global_config):

        BaseModel.__init__(self, model_config, global_config)

    def add_sitemodel(self, distribution, feature, fname):

            # Sitemodel
            if self.rate_variation:
                mr = "@featureClockRate:%s" % fname
            else:
                mr = "1.0"
            sitemodel = ET.SubElement(distribution, "siteModel", {"id":"geoSiteModel.%s"%fname,"spec":"SiteModel", "mutationRate":mr,"shape":"1","proportionInvariant":"0"})

            substmodel = ET.SubElement(sitemodel, "substModel",{"id":"ordinal.s:%s"%fname,"spec":"Ordinal","stateNumber":"%d" % self.valuecounts[feature]})

            # Empirical frequencies
            if self.frequencies == "empirical":
                if self.pruned:
                    freq = ET.SubElement(substmodel,"frequencies",{"id":"featurefreqs.s:%s"%fname,"spec":"Frequencies", "data":"@%s.filt"%fname})
                else:
                    freq = ET.SubElement(substmodel,"frequencies",{"id":"featurefreqs.s:%s"%fname,"spec":"Frequencies", "data":"@%s"%fname})



import sys
import collections
from .unicodecsv import UnicodeDictReader

def sniff_format(fp):
    header = fp.readline()
    # Is this a CLDF format?
    if (all([f in header for f in ("Language_ID", "Value")])
        and
        any([f in header for f in ("Feature_ID", "Parameter_ID")])
       ):
        diag = "cldf"
    # If not, assume it uses the default BEASTling format
    else:
        diag = "beastling"
    return diag, header

def load_data(filename, file_format=None, lang_column=None):
    # Open file
    if filename == "stdin":
        fp = sys.stdin
    else:
        fp = open(filename, "r")

    # Determine format (if not given)
    if not file_format:
        file_format, header = sniff_format(fp)
    else:
        header = fp.readline()

    # Load data
    if file_format == "cldf":
        data = load_cldf_data(fp, header)
    elif file_format == "beastling":
        data = load_beastling_data(fp, header, lang_column, filename)

    # Close file if necessary
    if filename != "stdin":
        fp.close()

    return data

_language_column_names = ("iso", "iso_code", "glotto", "glottocode", "language", "language_id", "lang", "lang_id")

def load_beastling_data(fp, header, lang_column, filename):
    reader = UnicodeDictReader(fp, header)
    if not lang_column:
        for candidate in reader.fieldnames:
            if candidate.lower() in _language_column_names:
                lang_column = candidate
                break

    if not lang_column or lang_column not in reader.fieldnames:
        raise ValueError("Cold not find language column in data file %s" % filename)
    data = collections.defaultdict(lambda: collections.defaultdict(lambda: "?"))
    for row in reader:
        if row[lang_column] in data:
            raise ValueError("Duplicated language identifier '%s' found in data file %s" % (row[lang_column], filename))
        data[row[lang_column]] = collections.defaultdict(lambda : "?", row)
    return data

def load_cldf_data(fp, header):
    if "Feature_ID" in header:
        feature_column = "Feature_ID"
    else:
        feature_column = "Parameter_ID"
    reader = UnicodeDictReader(fp, header)
    data = collections.defaultdict(lambda: collections.defaultdict(lambda: "?"))
    for row in reader:
        lang = row["Language_ID"]
        if lang not in data:
            data[lang] = collections.defaultdict(lambda :"?")
        data[lang][row[feature_column]] = row["Value"]
    return data



maps = (
    ("Beta", "beast.math.distributions.Beta"),
    ("Exponential", "beast.math.distributions.Exponential"),
    ("InverseGamma", "beast.math.distributions.InverseGamma"),
    ("LogNormal", "beast.math.distributions.LogNormalDistributionModel"),
    ("Gamma", "beast.math.distributions.Gamma"),
    ("Uniform", "beast.math.distributions.Uniform"),
    ("LaplaceDistribution", "beast.math.distributions.LaplaceDistribution"),
    ("OneOnX", "beast.math.distributions.OneOnX"),
    ("Normal", "beast.math.distributions.Normal"),
    ("prior", "beast.math.distributions.Prior"),
    ("taxon", "beast.evolution.alignment.Taxon"),
    ("taxonset", "beast.evolution.alignment.TaxonSet"),
)



import codecs
import datetime
import os
import pdb
import sys
import xml.etree.ElementTree as ET

from beastling import __version__
import beastling.beast_maps as beast_maps

def indent(elem, level=0):
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent(elem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i

class BeastXml:

    def __init__(self, config):
        self.config = config
        if not self.config.processed:
            self.config.process()
        self.build_xml()

    def build_xml(self):

        # Root "beast" node
        attribs = {}
        attribs["beautitemplate"] = "Standard"
        attribs["beautistatus"] = ""
        attribs["namespace"] = "beast.core:beast.evolution.alignment:beast.evolution.tree.coalescent:beast.core.util:beast.evolution.nuc:beast.evolution.operators:beast.evolution.sitemodel:beast.evolution.substitutionmodel:beast.evolution.likelihood"
        attribs["version"] ="2.0"
        self.beast = ET.Element("beast", attrib=attribs)

        # "Generated by..." comment
        comment_lines = []
        comment_lines.append("Generated by BEASTling %s on %s." % (__version__,datetime.datetime.now().strftime("%A, %d %b %Y %l:%M %p")))
        if self.config.configfile_text:
            comment_lines.append("Original config file:")
            comment_lines.append(self.config.configfile_text)
        else:
            comment_lines.append("Configuration built programmatically.")
            comment_lines.append("No config file to include.")
        self.beast.append(ET.Comment("\n".join(comment_lines)))

        # Maps
        for a, b in beast_maps.maps:
            mapp = ET.SubElement(self.beast, "map", attrib={"name":a})
            mapp.text = b

        # State
        state = ET.SubElement(self.beast, "state", {"id":"state","storeEvery":"5000"})

        ## Taxon set
        taxonset = ET.SubElement(state, "taxonset", {"id":"taxa"})
        for lang in self.config.languages:
            ET.SubElement(taxonset, "taxon", {"id":lang,})

        ## Tree
        tree = ET.SubElement(state, "tree", {"id":"Tree.t:beastlingTree"})
        ET.SubElement(tree, "taxonset", {"idref":"taxa"})

        ## Birth model
        param = ET.SubElement(state, "parameter", {"id":"birthRate.t:beastlingTree","name":"stateNode"})
        param.text="1.0"

        for model in self.config.models:
            model.add_state(state)

        # Misc
        for model in self.config.models:
            model.add_misc(self.beast)

        # Run
        attribs = {}
        attribs["id"] = "mcmc"
        attribs["spec"] = "MCMC"
        attribs["chainLength"] = str(self.config.chainlength)
        if self.config.sample_from_prior:
            attribs["sampleFromPrior"] = "true"
        self.run = ET.SubElement(self.beast, "run", attrib=attribs)

        ## Init
        ### If a starting tree is specified, use it...
        if self.config.starting_tree:
            init = ET.SubElement(self.run, "init", {"estimate":"false", "id":"startingTree", "initial":"@Tree.t:beastlingTree", "spec":"beast.util.TreeParser","newick":self.config.starting_tree})
        ### ...if not, use a random tree
        else:
            ### But the random tree must respect any constraints!
            if self.config.monophyly:
                init = ET.SubElement(self.run, "init", {"estimate":"false", "id":"startingTree", "initial":"@Tree.t:beastlingTree", "taxonset":"@taxa", "spec":"beast.evolution.tree.ConstrainedRandomTree", "constraints":"@constraints"})
            else:
                init = ET.SubElement(self.run, "init", {"estimate":"false", "id":"startingTree", "initial":"@Tree.t:beastlingTree", "taxonset":"@taxa", "spec":"beast.evolution.tree.RandomTree"})
            popmod = ET.SubElement(init, "populationModel", {"spec":"ConstantPopulation"})
            ET.SubElement(popmod, "popSize", {"spec":"parameter.RealParameter","value":"1"})

        ## Distributions
        self.master_distribution = ET.SubElement(self.run,"distribution",{"id":"posterior","spec":"util.CompoundDistribution"})

        ### Prior
        self.add_prior()
        for model in self.config.models:
            model.add_prior(self.prior)

        ### Likelihood
        self.likelihood = ET.SubElement(self.master_distribution,"distribution",{"id":"likelihood","spec":"util.CompoundDistribution"})
        for model in self.config.models:
            model.add_likelihood(self.likelihood)

        ## Operators
        self.add_operators()

        ## Logging
        self.add_loggers()

    def add_prior(self):

        """Add "master prior" features, independent of any data
        or models.  E.g. monophyly constraints, clade calibration
        dates, tree priors, etc."""

        self.prior = ET.SubElement(self.master_distribution,"distribution",{"id":"prior","spec":"util.CompoundDistribution"})

        # Monophyly
        glotto_iso_langs = [l for l in self.config.languages if l.lower() in self.config.classifications]
        if self.config.monophyly and glotto_iso_langs:
            attribs = {}
            attribs["id"] = "constraints"
            attribs["spec"] = "beast.math.distributions.MultiMonophyleticConstraint"
            attribs["tree"] = "@Tree.t:beastlingTree"
            attribs["newick"] = self.make_monophyly_newick(glotto_iso_langs)
            ET.SubElement(self.prior, "distribution", attribs)

        # Calibration
        if self.config.calibrations:
            for n, clade in enumerate(self.config.calibrations):
                if clade == "root":
                    langs = self.config.languages
                else:
                    langs = [l for l in self.config.languages if any([c==clade for c in [x.lower() for x in self.config.classifications[l.lower()].split(",")]])]
                if not langs:
                    continue
                lower, upper = self.config.calibrations[clade]
                mean = (upper + lower) / 2.0
                stddev = (upper - mean) / 2.0
                attribs = {}
                attribs["id"] = clade + "-calibration.prior"
                attribs["monophyletic"] = "true"
                attribs["spec"] = "beast.math.distributions.MRCAPrior"
                attribs["tree"] = "@Tree.t:beastlingTree"
                cal_prior = ET.SubElement(self.prior, "distribution", attribs)

                taxonset = ET.SubElement(cal_prior, "taxonset", {"id" : clade, "spec":"TaxonSet"})
                for lang in langs:
                    ET.SubElement(taxonset, "taxon", {"idref":lang})
                normal = ET.SubElement(cal_prior, "Normal", {"id":"CalibrationNormal.%d" % n, "name":"distr", "offset":str(mean)})
                ET.SubElement(normal, "parameter", {"id":"parameter.hyperNormal-mean-%s.prior" % clade, "name":"mean", "estimate":"false"}).text = "0.0"
                ET.SubElement(normal, "parameter", {"id":"parameter.hyperNormal-sigma-%s.prior" % clade, "name":"sigma", "estimate":"false"}).text = str(stddev)


        # Tree prior
        attribs = {}
        attribs["birthDiffRate"] = "@birthRate.t:beastlingTree"
        attribs["id"] = "YuleModel.t:beastlingTree"
        attribs["spec"] = "beast.evolution.speciation.YuleModel"
        attribs["tree"] = "@Tree.t:beastlingTree"
        ET.SubElement(self.prior, "distribution", attribs)

        # Birth rate
        attribs = {}
        attribs["id"] = "YuleBirthRatePrior.t:beastlingTree"
        attribs["name"] = "distribution"
        attribs["x"] = "@birthRate.t:beastlingTree"
        sub_prior = ET.SubElement(self.prior, "prior", attribs)
        uniform = ET.SubElement(sub_prior, "Uniform", {"id":"Uniform.0","name":"distr","upper":"Infinity"})


    def add_operators(self):

        # Tree operators
        if self.config.sample_branch_lengths:
            ## Branch length operators
            ET.SubElement(self.run, "operator", {"id":"UniformOperator.t:beastlingTree","spec":"Uniform","tree":"@Tree.t:beastlingTree","weight":"30.0"})
            ET.SubElement(self.run, "operator", {"id":"treeScaler.t:beastlingTree","scaleFactor":"1.0","spec":"ScaleOperator","tree":"@Tree.t:beastlingTree","weight":"3.0"})
            ET.SubElement(self.run, "operator", {"id":"treeRootScaler.t:beastlingTree","scaleFactor":"1.0","spec":"ScaleOperator","tree":"@Tree.t:beastlingTree","rootOnly":"true","weight":"3.0"})
        if self.config.sample_topology:
            ## Tree topology operators
            ET.SubElement(self.run, "operator", {"id":"SubtreeSlide.t:beastlingTree","spec":"SubtreeSlide","tree":"@Tree.t:beastlingTree","markclades":"true", "weight":"15.0"})
            ET.SubElement(self.run, "operator", {"id":"narrow.t:beastlingTree","spec":"Exchange","tree":"@Tree.t:beastlingTree","markclades":"true", "weight":"15.0"})
            ET.SubElement(self.run, "operator", {"id":"wide.t:beastlingTree","isNarrow":"false","spec":"Exchange","tree":"@Tree.t:beastlingTree","markclades":"true", "weight":"3.0"})
            ET.SubElement(self.run, "operator", {"id":"WilsonBalding.t:beastlingTree","spec":"WilsonBalding","tree":"@Tree.t:beastlingTree","markclades":"true","weight":"3.0"})

        # Birth rate scaler
        ET.SubElement(self.run, "operator", {"id":"YuleBirthRateScaler.t:beastlingTree","spec":"ScaleOperator","parameter":"@birthRate.t:beastlingTree", "scaleFactor":"1.0", "weight":"3.0"})

        # Up/down
        updown = ET.SubElement(self.run, "operator", {"id":"UpDown","spec":"UpDownOperator","scaleFactor":"1.0", "weight":"3.0"})
        ET.SubElement(updown, "tree", {"idref":"Tree.t:beastlingTree", "name":"up"})
        ET.SubElement(updown, "parameter", {"idref":"birthRate.t:beastlingTree", "name":"down"})
        if self.config.calibrations:
            for model in self.config.models:
                ET.SubElement(updown, "parameter", {"idref":"clockRate.c:%s" % model.name, "name":"down"})

        # Model specific operators
        for model in self.config.models:
            model.add_operators(self.run)

    def add_loggers(self):

        # Screen logger
        if self.config.screenlog:
            screen_logger = ET.SubElement(self.run, "logger", attrib={"id":"screenlog", "logEvery":str(self.config.log_every)})
            log = ET.SubElement(screen_logger, "log", attrib={"arg":"@posterior", "id":"ESS.0", "spec":"util.ESS"})
            log = ET.SubElement(screen_logger, "log", attrib={"idref":"prior"})
            log = ET.SubElement(screen_logger, "log", attrib={"idref":"likelihood"})
            log = ET.SubElement(screen_logger, "log", attrib={"idref":"posterior"})

        # Tracer log
        if self.config.log_probabilities or self.config.log_params or self.config.log_all:
            tracer_logger = ET.SubElement(self.run,"logger",{"id":"tracelog","fileName":self.config.basename+".log","logEvery":str(self.config.log_every),"model":"@posterior","sanitiseHeaders":"true","sort":"smart"})
            if self.config.log_probabilities or self.config.log_all:
                ET.SubElement(tracer_logger,"log",{"idref":"prior"})
                ET.SubElement(tracer_logger,"log",{"idref":"likelihood"})
                ET.SubElement(tracer_logger,"log",{"idref":"posterior"})
            if self.config.log_params or self.config.log_all:
                ET.SubElement(tracer_logger,"log",{"idref":"birthRate.t:beastlingTree"})
                for model in self.config.models:
                    model.add_param_logs(tracer_logger)
                
        # Tree log
        if self.config.log_trees or self.config.log_all:
            tree_logger = ET.SubElement(self.run, "logger", {"mode":"tree", "fileName":self.config.basename+".nex","logEvery":str(self.config.log_every),"id":"treeWithMetaDataLogger"})
            log = ET.SubElement(tree_logger, "log", attrib={"id":"TreeLogger","spec":"beast.evolution.tree.TreeWithMetaDataLogger","tree":"@Tree.t:beastlingTree"})

    def write_file(self, filename=None):
        indent(self.beast)
        xml_string = ET.tostring(self.beast, encoding="UTF-8")
        if not filename:
            filename = self.config.basename+".xml"
        if filename in ("stdout", "-"):
            sys.stdout.write(unicode(xml_string, "utf-8"))
        else:
            fp = codecs.open(filename, "w", "UTF-8")
            fp.write(unicode(xml_string, "utf-8"))
            fp.close()

    def make_tight_monophyly_structure(self, langs, depth=0, maxdepth=sys.maxint):
        if depth > maxdepth:
            return langs
        levels = list(set([self.config.classifications[l.lower()].split(",")[depth] for l in langs]))
        if len(levels) == 1:
            if levels[0] == "":
                langs.sort()
                return langs
            else:
                return self.make_tight_monophyly_structure(langs, depth+1, maxdepth)
        else:
            partition = [[l for l in langs if self.config.classifications[l.lower()].split(",")[depth] == level] for level in levels]
            partition = [part for part in partition if part]
            return sorted([self.make_tight_monophyly_structure(group, depth+1, maxdepth) for group in partition])

    def make_loose_monophyly_structure(self, langs):
        points = self.config.monophyly
        return [[l for l in langs if point in self.config.classifications[l.lower()] ] for point in points]

    def make_monophyly_string(self, struct, depth=0):
        if not type([]) in [type(x) for x in struct]:
            return "(%s)" % ",".join(struct)
        else:
            return "(%s)" % ",".join([self.make_monophyly_string(substruct) for substruct in struct])

    def make_monophyly_newick(self, langs):
        if self.config.monophyly_grip == "tight":
            struct = self.make_tight_monophyly_structure(langs, depth=self.config.monophyly_start_depth, maxdepth=self.config.monophyly_end_depth)
        elif self.config.monophyly_grip == "loose":
            struct = self.make_loose_monophyly_structure(langs)
        return self.make_monophyly_string(struct)




import codecs
import os
import xml.etree.ElementTree as ET

from .basemodel import BaseModel
from ..fileio.unicodecsv import UnicodeDictReader

class MKModel(BaseModel):

    package_notice = """[DEPENDENCY]: The Lewis Mk substitution model is implemented in the BEAST package "morph-models"."""

    def __init__(self, model_config, global_config):

        BaseModel.__init__(self, model_config, global_config)

    def add_sitemodel(self, distribution, feature, fname):

            # Sitemodel
            if self.rate_variation:
                mr = "@featureClockRate:%s" % fname
            else:
                mr = "1.0"
            sitemodel = ET.SubElement(distribution, "siteModel", {"id":"SiteModel.%s"%fname,"spec":"SiteModel", "mutationRate":mr,"shape":"1","proportionInvariant":"0"})

            substmodel = ET.SubElement(sitemodel, "substModel",{"id":"mk.s:%s"%fname,"spec":"LewisMK","datatype":"@traitDataType.%s" % fname})
            # Do empirical frequencies
            # We don't need to do anything for uniform freqs
            # as the implementation of LewisMK handles it
            if self.frequencies == "empirical":
                if self.pruned:
                    freq = ET.SubElement(substmodel,"frequencies",{"id":"featurefreqs.s:%s"%fname,"spec":"Frequencies", "data":"@%s.filt"%fname})
                else:
                    freq = ET.SubElement(substmodel,"frequencies",{"id":"featurefreqs.s:%s"%fname,"spec":"Frequencies", "data":"@%s"%fname})



import codecs
import ConfigParser
import csv
import itertools
import pkgutil
import os
import sys

import beastling.models.bsvs as bsvs
import beastling.models.covarion as covarion
import beastling.models.mk as mk

def assert_compare_equal(one, other):
    """ Compare two values. If they match, return that value, otherwise raise an error. """
    if one != other:
        raise ValueError("Values {:s} and {:s} were expected to match.".format(one, other))
    return one

class Configuration:
    valid_overlaps = {
        "union": set.union,
        "intersection": set.intersection,
        "error": assert_compare_equal}

    def __init__(self, basename="beastling", configfile=None, stdin_data=False):

        self.processed = False
        self.messages = []
        self.message_flags = []

        # Set up default options
        self.basename = basename
        self.configfile = None
        self.configfile_text = None
        self.chainlength = 10000000
        self.sample_from_prior = False
        self.families = "*"
        self.overlap = "error"
        self.starting_tree = ""
        self.sample_branch_lengths = True
        self.sample_topology = True
        self.model_configs = []
        self.monophyly = False
        self.monophyly_start_depth = 0
        self.monophyly_end_depth = sys.maxint
        self.monophyly_grip = "tight"
        self.screenlog = True
        self.log_all = False
        self.log_every = 0
        self.log_params = False
        self.log_probabilities = True
        self.log_trees = True
        self.stdin_data = stdin_data
        self.calibrations = {}

        if configfile:
            self.read_from_file(configfile)

    def read_from_file(self, configfile):
        # Read config file and overwrite defaults
        self.configfile = configfile
        fp = open(self.configfile, "r")
        self.configfile_text = fp.read()
        fp.close()
        p = ConfigParser.SafeConfigParser()
        p.read(self.configfile)

        ## Admin
        sec = "admin"
        if p.has_option(sec, "basename"):
            self.basename = p.get(sec, "basename")
        if p.has_option(sec, "screenlog"):
            self.screenlog = p.getboolean(sec, "screenlog")
        if p.has_option(sec, "log_every"):
            self.log_every = p.getint(sec, "log_every")
        if p.has_option(sec, "log_all"):
            self.log_all = p.getboolean(sec, "log_all")
        if p.has_option(sec, "log_probabilities"):
            self.log_probabilities = p.getboolean(sec, "log_probabilities")
        if p.has_option(sec, "log_params"):
            self.log_params = p.getboolean(sec, "log_params")
        if p.has_option(sec, "log_trees"):
            self.log_trees = p.getboolean(sec, "log_trees")
            
        ## MCMC
        sec = "MCMC"
        if p.has_option(sec, "chainlength"):
            self.chainlength = p.getint(sec, "chainlength")
        if p.has_option(sec, "sample_from_prior"):
            self.sample_from_prior = p.getboolean(sec, "sample_from_prior")

        ## Languages
        sec = "languages"
        if p.has_option(sec, "families"):
            self.families = p.get(sec, "families")
        if p.has_option(sec, "overlap"):
            self.overlap = p.get(sec, "overlap")
            if not self.overlap in Configuration.valid_overlaps:
                raise ValueError("Value for overlap needs to be one of 'union', 'intersection' or 'error'.")
                
        if p.has_option(sec, "starting_tree"):
            self.starting_tree = p.get(sec, "starting_tree")
        if p.has_option(sec, "sample_branch_lengths"):
            self.sample_branch_lengths = p.getboolean(sec, "sample_branch_lengths")
        if p.has_option(sec, "sample_topology"):
            self.sample_topology = p.getboolean(sec, "sample_topology")
        if p.has_option(sec, "monophyletic"):
            self.monophyly = p.getboolean(sec, "monophyletic")
        elif p.has_option(sec, "monophyly"):
            self.monophyly = p.getboolean(sec, "monophyly")
        if p.has_option(sec, "monophyly_start_depth"):
            self.monophyly_start_depth = p.getint(sec, "monophyly_start_depth")
        if p.has_option(sec, "monophyly_end_depth"):
            self.monophyly_end_depth = p.getint(sec, "monophyly_end_depth") - 1
        if p.has_option(sec, "monophyly_grip"):
            self.monophyly_grip = p.get(sec, "monophyly_grip").lower()

        ## Calibration
        if p.has_section("calibration"):
            for clade, dates in p.items("calibration"):
                self.calibrations[clade] = [float(x.strip()) for x in dates.split("-")]
            
        ## Models
        sections = p.sections()
        model_sections = [s for s in sections if s.lower().startswith("model")]
        if not model_sections:
            raise ValueError("Config file contains no model sections.")
        for section in model_sections:
            options = p.options(section)
            config = {option:p.get(section, option) for option in options}
            if "rate_variation" in config:
                config["rate_variation"] = p.getboolean(section,"rate_variation")
            else:
                config["rate_variation"] = False
            if "remove_constant_features" in config:
                config["remove_constant_features"] = p.getboolean(section,"remove_constant_features")
            else:
                config["remove_constant_features"] = True
            if "file_format" in config:
                config["file_format"] = p.getboolean(section,"file_format")
            if "language_column" in config:
                config["language_column"] = p.get(section,"language_column")
            config["name"] = section[5:].strip() # Chop off "model" prefix
            self.model_configs.append(config)

    def load_glotto_class(self):
        self.classifications = {}
        glotto_data = pkgutil.get_data('beastling', 'data/glotto.csv')
        glotto_classifications = self.parse_glotto_class(glotto_data)
        iso_glotto_data = pkgutil.get_data('beastling', 'data/iso-glotto.csv')
        iso_classifications = self.parse_glotto_class(iso_glotto_data)
        for key in glotto_classifications:
            self.classifications[key] = glotto_classifications[key]
        for key in iso_classifications:
            self.classifications[key] = iso_classifications[key]

    def parse_glotto_class(self, binary_glotto):
        classifications = {}
        unicode_glotto = unicode(binary_glotto, "UTF-8")
        glotto_lines = unicode_glotto.split("\n")
        glotto_lines = [l for l in glotto_lines if l]
        for line in glotto_lines:
            lang, clazz = line.split(",",1)
            lang = lang.strip().lstrip().lower()
            if clazz.lstrip().strip() == "Unclassified":
                continue

            clazz = clazz.split(",")
            if not clazz[0]:
                clazz[0] = "Isolate_" + lang
            clazz = ",".join(clazz)
            classifications[lang] = clazz
        return classifications

    def process(self):

        # Add dependency notice if required
        if self.monophyly and not self.starting_tree:
            self.messages.append("[DEPENDENCY] ConstrainedRandomTree is implemented in the BEAST package BEASTLabs.")

        # If log_every was not explicitly set to some non-zero
        # value, then set it such that we expect 10,000 log
        # entries
        if not self.log_every:
            self.log_every = self.chainlength / 10000
            ## If chainlength < 10000, this results in log_every = zero.
            ## This causes BEAST to die.
            ## So in this case, just log everything.
            if self.log_every == 0:
                self.log_every = 1

        if os.path.exists(self.families):
            fp = codecs.open(self.families, "r", "UTF-8")
            self.families = [x.strip() for x in fp.readlines()]
            fp.close()
        else:
            self.families = [x.strip() for x in self.families.split(",")]

        # Read starting tree from file
        if os.path.exists(self.starting_tree):
            fp = codecs.open(self.starting_tree, "r", "UTF-8")
            self.starting_tree = fp.read().strip()
            fp.close()

        ## Load Glottolog classifications
        self.load_glotto_class()

        ## Determine final list of languages
        if self.families == ["*"]:
            self.lang_filter = []
        else:
            self.lang_filter = [l for l in self.classifications if any([family in self.classifications[l] for family in self.families])]

        # Hack to fix Glottolog's broken Bontok
        if "bnc" in self.lang_filter:
            self.lang_filter.remove("bnc")
            self.lang_filter.append("lbk")

        # Handle request to read data from stdin
        if self.stdin_data:
            for config in self.model_configs:
                config["data"] = "stdin"
        # Instantiate models
        if not self.model_configs:
            raise ValueError("No models specified!")
        self.models = []
        for config in self.model_configs:
            if "model" not in config:
                raise ValueError("Model not specified for model section %s." % config["name"])
            if "data" not in config:
                raise ValueError("Data source not specified in model section %s." % config["name"])
            if config["model"].lower() == "bsvs":
                model = bsvs.BSVSModel(config, self)
                if "bsvs_used" not in self.message_flags:
                    self.message_flags.append("bsvs_used")
                    self.messages.append(bsvs.BSVSModel.package_notice)
            elif config["model"].lower() == "covarion":
                model = covarion.CovarionModel(config, self)
            elif config["model"].lower() == "mk":
                model = mk.MKModel(config, self)
                if "mk_used" not in self.message_flags:
                    self.message_flags.append("mk_used")
                    self.messages.append(mk.MKModel.package_notice)
            else:
                raise ValueError("Unknown model type '%s' for model section '%s'." % (config["model"], config["name"]))
            if config["model"].lower() != "covarion":
                self.messages.append("""[DEPENDENCY] Model %s: AlignmentFromTrait is implemented in the BEAST package "BEAST_CLASSIC".""" % config["name"])
            self.messages.extend(model.messages)
            self.models.append(model)

        # Finalise language list.
        ## Start with all the languages from a random data source
        self.languages = set(self.models[0].data.keys())
        overlap_resolver = Configuration.valid_overlaps[self.overlap]
        for model in self.models:
            # A filter is just a set.
            if self.lang_filter:
                addition = set(model.data.keys()) & self.lang_filter
            else:
                addition = set(model.data.keys())
            # This depends on the value of `overlap`.
            self.languages = overlap_resolver(self.languages, addition)

        ## Apply family-based filtering
        ## Make sure there's *something* left
        if not self.languages:
            raise ValueError("No languages specified!")

        ## Convert back into a sorted list
        self.languages = sorted(self.languages)
        self.messages.append("[INFO] %d languages included in analysis." % len(self.languages))

        self.processed = True

if __name__ == "__main__":
    main()



import codecs
import os
import xml.etree.ElementTree as ET

from .basemodel import BaseModel
from ..fileio.unicodecsv import UnicodeDictReader

class CovarionModel(BaseModel):

    def __init__(self, model_config, global_config):

        BaseModel.__init__(self, model_config, global_config)
        self.freq_str = self.build_freq_str()

    def build_freq_str(self):
        all_data = []
        for n, f in enumerate(self.features):
            fname = "%s:%s" % (self.name, f)
            frange = sorted(list(set(self.data[lang][f] for lang in self.data)))
            for lang in self.data:
                if self.data[lang].get(f,"?") == "?":
                    valuestring = "".join(["?" for i in range(0,len(frange)+1)])
                else:
                    valuestring = ["0" for i in range(0,len(frange)+1)]
                    valuestring[frange.index(self.data[lang][f])+1] = "1"
                    all_data.extend(valuestring)

        all_data = [d for d in all_data if d !="?"]
        zerf = 1.0*all_data.count("0") / len(all_data)
        onef = 1.0*all_data.count("1") / len(all_data)
        return "%.2f %.2f" % (zerf, onef)

    def add_state(self, state):
        BaseModel.add_state(self, state)
        alpha = ET.SubElement(state, "parameter", {"id":"%s:covarion_alpha.s" % self.name, "lower":"1.0E-4", "name":"stateNode", "upper":"1.0"})
        alpha.text="0.5"
        switch = ET.SubElement(state, "parameter", {"id":"%s:covarion_s.s" % self.name, "lower":"1.0E-4", "name":"stateNode", "upper":"Infinity"})
        switch.text="0.5"

        vfreq = ET.SubElement(state, "parameter", {"id":"%s:visiblefrequencies.s" % self.name, "dimension":"2", "lower": "0.0", "upper":"1.0", "name":"stateNode"})
        vfreq.text="0.5 0.5"

    def add_data(self, distribution, feature, fname):
        frange = sorted(list(set(self.data[lang][feature] for lang in self.config.languages)))
        data = ET.SubElement(distribution,"data",{"id":fname, "spec":"Alignment", "ascertained":"true", "excludefrom":"0","excludeto":"1"})
        ET.SubElement(data, "userDataType",{"spec":"beast.evolution.datatype.TwoStateCovarion"})
        for lang in self.config.languages:
            if self.data[lang][feature] == "?":
                valuestring = "".join(["?" for i in range(0,len(frange)+1)])
            else:
                valuestring = ["0" for i in range(0,len(frange)+1)]
                valuestring[frange.index(self.data[lang][feature])+1] = "1"
                valuestring = "".join(valuestring)

            seq = ET.SubElement(data, "sequence", {"id":"seq_%s_%s" % (lang, fname), "taxon":lang, "totalcount":"4","value":valuestring})

    def add_misc(self, beast):
        # The "vfrequencies" parameter here is the frequencies
        # of the *visible* states (present/absent) and should
        # be based on the data (if we are doing an empirical
        # analysis)
        substmodel = ET.SubElement(beast, "substModel",{"id":"covarion.s","spec":"BinaryCovarion","alpha":"@%s:covarion_alpha.s" % self.name, "switchRate":"@%s:covarion_s.s" % self.name})
        vfreq = ET.SubElement(substmodel, "parameter", {"id":"@%s:visiblefrequencies.s" % self.name,"dimension":"2","name":"vfrequencies"})
        if self.frequencies == "empirical":
            vfreq.text = self.freq_str
        elif self.frequencies == "uniform":
            vfreq.text="0.5 0.5"
        # These are the frequencies of the *hidden* states
        # (fast / slow), and are just set to 50:50 
        hfreq = ET.SubElement(substmodel, "parameter", {"id":"hiddenfrequencies.s","dimension":"2","lower":"0.0","name":"hfrequencies","upper":"1.0"})
        hfreq.text="0.5 0.5"

        # Dummy frequencies - these do nothing and are required
        # to stop the BinaryCovarion model complaining that the
        # "frequencies" input is not specified, which is
        # inherited behaviour from GeneralSubstitutionModel
        # which probably should have been overridden...
        freq = ET.SubElement(substmodel, "frequencies", {"id":"dummyfrequences.s","spec":"Frequencies","frequencies":"0.5 0.5"})

    def add_sitemodel(self, distribution, feature, fname):

        # Sitemodel
        if self.rate_variation:
            mr = "@featureClockRate:%s" % fname
        else:
            mr = "1.0"
        sitemodel = ET.SubElement(distribution, "siteModel", {"id":"SiteModel.%s"%fname,"spec":"SiteModel", "mutationRate":mr,"shape":"1","proportionInvariant":"0", "substModel":"@covarion.s"})

    def add_prior(self, prior):
        BaseModel.add_prior(self, prior)
        alpha_prior = ET.SubElement(prior, "prior", {"id":"covarion_alpha_prior.s","name":"distribution","x":"@%s:covarion_alpha.s" % self.name})
        ET.SubElement(alpha_prior, "Uniform", {"id":"CovAlphaUniform","name":"distr","upper":"Infinity"})
        switch_prior = ET.SubElement(prior, "prior", {"id":"covarion_s_prior.s","name":"distribution","x":"@%s:covarion_s.s" % self.name})
        gamma = ET.SubElement(switch_prior, "Gamma", {"id":"Gamma.0", "name":"distr"})
        ET.SubElement(gamma, "parameter", {"id":"covarion_switch_gamma_param1","name":"alpha","lower":"0.0","upper":"0.0"}).text = "0.05"
        ET.SubElement(gamma, "parameter", {"id":"covarion_switch_gamma_param2","name":"beta","lower":"0.0","upper":"0.0"}).text = "10.0"

    def add_operators(self, run):
        BaseModel.add_operators(self, run)
        ET.SubElement(run, "operator", {"id":"covarion_alpha_scaler.s", "spec":"ScaleOperator","parameter":"@%s:covarion_alpha.s" % self.name,"scaleFactor":"0.75","weight":"0.1"})
        ET.SubElement(run, "operator", {"id":"covarion_s_scaler.s", "spec":"ScaleOperator","parameter":"@%s:covarion_s.s" % self.name,"scaleFactor":"0.75","weight":"0.1"})
        delta = ET.SubElement(run, "operator", {"id":"frequenciesDelta", "spec":"DeltaExchangeOperator","delta":"0.01","weight":"0.1"})
        ET.SubElement(delta, "parameter", {"idref":"%s:visiblefrequencies.s" % self.name})

    def add_param_logs(self, logger):
        BaseModel.add_param_logs(self, logger)
        ET.SubElement(logger,"log",{"idref":"%s:covarion_alpha.s" % self.name})
        ET.SubElement(logger,"log",{"idref":"%s:covarion_s.s" % self.name})



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from beastling import __version__ as version

setup(
    name='beastling',
    version=version,
    description='Command line tool to help mortal linguists use BEAST',
    author='Luke Maurits',
    author_email='luke@maurits.id.au',
    license="BSD (3 clause)",
    classifiers=[
        'Programming Language :: Python',
        'License :: OSI Approved :: BSD License',
    ],
    packages=['beastling','beastling.fileio','beastling.models'],
    package_data={'beastling': ['data/*']},
    scripts=['bin/beastling',],
)



import codecs
import os
import xml.etree.ElementTree as ET

import scipy.stats

from ..fileio.datareaders import load_data, _language_column_names

class BaseModel:

    def __init__(self, model_config, global_config):

        self.messages = []
        self.config = global_config
        self.calibrations = global_config.calibrations

        self.name = model_config["name"] 
        self.data_filename = model_config["data"] 
        if "features" in model_config:
            self.features = model_config["features"] 
        else:
            self.features = "*"
        self.constant_feature = False
        self.frequencies = model_config.get("frequencies", "empirical")
        self.pruned = model_config.get("pruned", False)
        self.rate_variation = model_config.get("rate_variation", False)
        self.remove_constant_features = model_config.get("remove_constant_features", True)
        self.lang_column = model_config.get("language_column", None)

        self.data = load_data(self.data_filename, file_format=model_config.get("file_format",None), lang_column=model_config.get("language_column",None))
        self.load_features()
        self.preprocess()

    def build_codemap(self, unique_values):
        N = len(unique_values)
        codemapbits = []
        codemapbits.append(",".join(["%s=%d" % (v,n) for (n,v) in enumerate(unique_values)]))
        codemapbits.append("?=" + " ".join([str(n) for n in range(0,N)]))
        codemapbits.append("-=" + " ".join([str(n) for n in range(0,N)]))
        return ",".join(codemapbits)

    def preprocess(self):

        # Remove features which are in the config but not the
        # data file
        self.features = [f for f in self.features if
                any([f in self.data[lang] for lang in self.data]
                    )]

        self.valuecounts = {}
        self.counts = {}
        self.dimensions = {}
        self.codemaps = {}
        bad_feats = []
        for f in self.features:
            all_values = [self.data[l][f] for l in self.data]
            all_values = [v for v in all_values if v != "?"]
            uniq = list(set(all_values))
            counts = {}
            for v in all_values:
                counts[v] = all_values.count(v)
            uniq = list(set(all_values))
            # Sort uniq carefully.
            # Possibly all feature values are numeric strings, e.g. "1", "2", "3".
            # If we sort these as strings then we get weird things like "10" < "2".
            # This can actually matter for things like ordinal models.
            # So convert these to ints first...
            if all([v.isdigit() for v in uniq]):
                uniq = map(int, uniq)
                uniq.sort()
                uniq = map(str, uniq)
            # ...otherwise, just sort normally
            else:
                uniq.sort()
            if len(uniq) == 0:
                self.messages.append("""[INFO] Model "%s": Feature %s excluded because there are no datapoints for selected languages.""" % (self.name, f))
                bad_feats.append(f)
                continue
            if len(uniq) == 1:
                if self.remove_constant_features:
                    self.messages.append("""[INFO] Model "%s": Feature %s excluded because its value is constant across selected languages.  Set "remove_constant_features=False" in config to stop this.""" % (self.name, f))
                    bad_feats.append(f)
                    continue
                else:
                    self.constant_feature = True
            N = len(uniq)

            self.valuecounts[f] = N
            self.counts[f] = counts
            self.dimensions[f] = N*(N-1)/2
            self.codemaps[f] = self.build_codemap(uniq)
        self.features = [f for f in self.features if f not in bad_feats]
        self.features.sort()
        self.messages.append("""[INFO] Model "%s": Using %d features from data source %s""" % (self.name, len(self.features), self.data_filename))
        if self.constant_feature and self.rate_variation:
            self.messages.append("""[WARNING] Model "%s": Rate variation enabled with constant features retained in data.  This may skew rate estimates for non-constant features.""" % self.name)
        if self.pruned:
            self.messages.append("""[DEPENDENCY] Model %s: Pruned trees are implemented in the BEAST package "BEASTlabs".""" % self.name)

    def load_features(self):
        # Load features to analyse
        if os.path.exists(self.features):
            features = []
            fp = codecs.open(self.features, "r", "UTF-8")
            for line in fp:
                feature = line.strip()
                features.append(feature)
        elif self.features == "*":
            random_iso = self.data.keys()[0]
            features = self.data[random_iso].keys()
            # Need to remove the languge ID column
            if self.lang_column:
                features.remove(self.lang_column)
            else:
                # If no language column name was explicitly given, just
                # remove the first of the automatically-recognised names
                # which we encounter:
                for lc in _language_column_names:
                    if lc in features:
                        features.remove(lc)
                        break
        else:
            features = [f.strip() for f in self.features.split(",")]
        self.features = features

    def add_misc(self, beast):
        pass

    def add_state(self, state):

        # Clock
        attribs = {}
        attribs["id"] = "clockRate.c:%s" % self.name
        attribs["name"] = "stateNode"
        parameter = ET.SubElement(state, "parameter", attribs)
        parameter.text="1.0"

        # Mutation rates
        if self.rate_variation:
            for f in self.features:
                fname = "%s:%s" % (self.name, f)

                attribs = {}
                attribs["id"] = "featureClockRate:%s" % fname
                attribs["name"] = "stateNode"
                parameter = ET.SubElement(state, "parameter", attribs)
                parameter.text="1.0"
            parameter = ET.SubElement(state, "parameter", {"id":"featureClockRateGammaShape:%s" % self.name, "name":"stateNode"})
            parameter.text="2.0"

    def add_prior(self, prior):

        # Clock
        sub_prior = ET.SubElement(prior, "prior", {"id":"clockPrior:%s" % self.name, "name":"distribution","x":"@clockRate.c:%s" % self.name})
        uniform = ET.SubElement(sub_prior, "Uniform", {"id":"UniformClockPrior:%s" % self.name, "name":"distr", "upper":"Infinity"})

        # Mutation rates
        if self.rate_variation:
            sub_prior = ET.SubElement(prior, "prior", {"id":"featureClockRatePrior.s:%s" % self.name, "name":"distribution"})
            compound = ET.SubElement(sub_prior, "input", {"id":"featureClockRateCompound:%s" % self.name, "spec":"beast.core.parameter.CompoundValuable", "name":"x"})
            for f in self.features:
                fname = "%s:%s" % (self.name, f)
                var = ET.SubElement(compound, "var", {"idref":"featureClockRate:%s" % fname})
            gamma  = ET.SubElement(sub_prior, "input", {"id":"featureClockRatePriorGamma:%s" % self.name, "spec":"beast.math.distributions.SingleParamGamma", "name":"distr", "alpha":"@featureClockRateGammaShape:%s" % self.name})

            sub_prior = ET.SubElement(prior, "prior", {"id":"featureClockRateGammaShapePrior.s:%s" % self.name, "name":"distribution", "x":"@featureClockRateGammaShape:%s" % self.name})
            exp = ET.SubElement(sub_prior, "Exponential", {"id":"featureClockRateGammaShapePriorExponential.s:%s" % self.name, "name":"distr"})
            param = ET.SubElement(exp, "parameter", {"id":"featureClockRateGammaShapePriorParam:%s" % self.name, "name":"mean", "lower":"0.0", "upper":"0.0"})
            param.text = "1.0"

    def add_data(self, distribution, feature, fname):
        # Data
        if self.pruned:
            data = ET.SubElement(distribution,"data",{"id":"%s.filt" % fname, "spec":"PrunedAlignment"})
            source = ET.SubElement(data,"source",{"id":fname,"spec":"AlignmentFromTrait"})
            parent = source
        else:
            data = ET.SubElement(distribution,"data",{"id":fname, "spec":"AlignmentFromTrait"})
            parent = data
        traitset = ET.SubElement(parent, "traitSet", {"id":"traitSet.%s" % fname,"spec":"beast.evolution.tree.TraitSet","taxa":"@taxa","traitname":"discrete"})
        stringbits = []
        for lang in self.config.languages:
           if lang in self.data:
               stringbits.append("%s=%s," % (lang, self.data[lang][feature]))
           else:
               stringbits.append("%s=?," % lang)
        traitset.text = " ".join(stringbits)
        userdatatype = ET.SubElement(parent, "userDataType", {"id":"traitDataType.%s"%fname,"spec":"beast.evolution.datatype.UserDataType","codeMap":self.codemaps[feature],"codelength":"-1","states":str(self.valuecounts[feature])})


    def add_likelihood(self, likelihood):

        for n, f in enumerate(self.features):
            fname = "%s:%s" % (self.name, f)
            distribution = ET.SubElement(likelihood, "distribution",{"id":"traitedtreeLikelihood.%s" % fname,"spec":"TreeLikelihood","useAmbiguities":"true"})

            # Tree
            if self.pruned:
                tree = ET.SubElement(distribution, "tree", {"id":"@Tree.t:beastlingTree.%s" % fname, "spec":"beast.evolution.tree.PrunedTree","quickshortcut":"true","assert":"false"})
                ET.SubElement(tree, "tree", {"idref":"Tree.t:beastlingTree"})
                ET.SubElement(tree, "alignment", {"idref":"%s.filt"%fname})
            else:
                tree = ET.SubElement(distribution, "tree", {"idref":"Tree.t:beastlingTree", "spec":"beast.evolution.tree.Tree"})

            # Sitemodel
            self.add_sitemodel(distribution, f, fname)

            # Branchrate
            branchrate = ET.SubElement(distribution, "branchRateModel", {"id":"StrictClockModel.c:%s"%fname,"spec":"beast.evolution.branchratemodel.StrictClockModel","clock.rate":"@clockRate.c:%s" % self.name})
            
            # Data
            self.add_data(distribution, f, fname)

    def add_operators(self, run):

        # Clock scaler (only for calibrated analyses)
        if self.calibrations:
            ET.SubElement(run, "operator", {"id":"clockScaler.c:%s" % self.name, "spec":"ScaleOperator","parameter":"@clockRate.c:%s" % self.name, "scaleFactor":"1.0","weight":"3.0"})

        # Mutation rates
        if self.rate_variation:
            delta = ET.SubElement(run, "operator", {"id":"featureClockRateDeltaExchanger:%s" % self.name, "spec":"DeltaExchangeOperator", "weight":"3.0"})
            for f in self.features:
                fname = "%s:%s" % (self.name, f)
                param = ET.SubElement(delta, "parameter", {"idref":"featureClockRate:%s" % fname})
            
            ET.SubElement(run, "operator", {"id":"featureClockRateGammaShapeScaler:%s" % self.name, "spec":"ScaleOperator","parameter":"@featureClockRateGammaShape:%s" % self.name, "scaleFactor":"1.0","weight":"0.1"})

    def add_param_logs(self, logger):

        # Clock
        ET.SubElement(logger,"log",{"idref":"clockRate.c:%s" % self.name})

        # Mutation rates
        if self.rate_variation:
            for f in self.features:
                fname = "%s:%s" % (self.name, f)
                ET.SubElement(logger,"log",{"idref":"featureClockRate:%s" % fname})
            ET.SubElement(logger,"log",{"idref":"featureClockRateGammaShape:%s" % self.name})
